@@ -1,8 +1,3 @@
-''' Nose test generators
-
-Need function load / save / roundtrip tests
-
-'''
 import os
 from collections import OrderedDict
 from os.path import join as pjoin, dirname
@@ -22,7 +17,7 @@ from pytest import raises as assert_raises
 
 import numpy as np
 from numpy import array
-import scipy.sparse as SP
+from scipy.sparse import issparse, eye_array, coo_array, csc_array
 
 import scipy.io
 from scipy.io.matlab import MatlabOpaque, MatlabFunction, MatlabObject
@@ -38,6 +33,7 @@ from scipy._lib._util import VisibleDeprecationWarning
 
 
 test_data_path = pjoin(dirname(__file__), 'data')
+pytestmark = pytest.mark.thread_unsafe
 
 
 def mlarr(*args, **kwargs):
@@ -76,14 +72,14 @@ case_table4.append(
 case_table4.append(
     {'name': 'sparse',
      'classes': {'testsparse': 'sparse'},
-     'expected': {'testsparse': SP.coo_matrix(A)},
+     'expected': {'testsparse': coo_array(A)},
      })
 B = A.astype(complex)
 B[0,0] += 1j
 case_table4.append(
     {'name': 'sparsecomplex',
      'classes': {'testsparsecomplex': 'sparse'},
-     'expected': {'testsparsecomplex': SP.coo_matrix(B)},
+     'expected': {'testsparsecomplex': coo_array(B)},
      })
 case_table4.append(
     {'name': 'multi',
@@ -211,12 +207,12 @@ case_table5.append(
 case_table5.append(
     {'name': 'sparse',
      'classes': {'testsparse': 'sparse'},
-     'expected': {'testsparse': SP.coo_matrix(A)},
+     'expected': {'testsparse': coo_array(A)},
      })
 case_table5.append(
     {'name': 'sparsecomplex',
      'classes': {'testsparsecomplex': 'sparse'},
-     'expected': {'testsparsecomplex': SP.coo_matrix(B)},
+     'expected': {'testsparsecomplex': coo_array(B)},
      })
 case_table5.append(
     {'name': 'bool',
@@ -251,8 +247,8 @@ def types_compatible(var1, var2):
 
 def _check_level(label, expected, actual):
     """ Check one level of a potentially nested array """
-    if SP.issparse(expected):  # allow different types of sparse matrices
-        assert_(SP.issparse(actual))
+    if issparse(expected):  # allow different types of sparse matrices
+        assert_(issparse(actual))
         assert_array_almost_equal(actual.toarray(),
                                   expected.toarray(),
                                   err_msg=label,
@@ -263,8 +259,7 @@ def _check_level(label, expected, actual):
             f"Expected type {type(expected)}, got {type(actual)} at {label}")
     # A field in a record array may not be an ndarray
     # A scalar from a record array will be type np.void
-    if not isinstance(expected,
-                      (np.void, np.ndarray, MatlabObject)):
+    if not isinstance(expected, np.void | np.ndarray | MatlabObject):
         assert_equal(expected, actual)
         return
     # This is an ndarray-like thing
@@ -295,11 +290,11 @@ def _check_level(label, expected, actual):
 
 def _load_check_case(name, files, case):
     for file_name in files:
-        matdict = loadmat(file_name, struct_as_record=True)
+        matdict = loadmat(file_name, struct_as_record=True, spmatrix=False)
         label = f"test {name}; file {file_name}"
         for k, expected in case.items():
             k_label = f"{label}, variable {k}"
-            assert_(k in matdict, "Missing key at %s" % k_label)
+            assert_(k in matdict, f"Missing key at {k_label}")
             _check_level(k_label, expected, matdict[k])
 
 
@@ -377,7 +372,7 @@ def test_gzip_simple():
     xdense = np.zeros((20,20))
     xdense[2,3] = 2.3
     xdense[4,5] = 4.5
-    x = SP.csc_matrix(xdense)
+    x = csc_array(xdense)
 
     name = 'gzip_test'
     expected = {'x':x}
@@ -391,7 +386,7 @@ def test_gzip_simple():
         mat_stream.close()
 
         mat_stream = gzip.open(fname, mode='rb')
-        actual = loadmat(mat_stream, struct_as_record=True)
+        actual = loadmat(mat_stream, struct_as_record=True, spmatrix=False)
         mat_stream.close()
     finally:
         shutil.rmtree(tmpdir)
@@ -1001,7 +996,7 @@ def test_mat_dtype():
 def test_sparse_in_struct():
     # reproduces bug found by DC where Cython code was insisting on
     # ndarray return type, but getting sparse matrix
-    st = {'sparsefield': SP.coo_matrix(np.eye(4))}
+    st = {'sparsefield': eye_array(4)}
     stream = BytesIO()
     savemat(stream, {'a':st})
     d = loadmat(stream, struct_as_record=True)
@@ -1162,9 +1157,9 @@ def test_logical_sparse():
     filename = pjoin(test_data_path,'logical_sparse.mat')
     # Before fix, this would crash with:
     # ValueError: indices and data should have the same size
-    d = loadmat(filename, struct_as_record=True)
+    d = loadmat(filename, struct_as_record=True, spmatrix=False)
     log_sp = d['sp_log_5_4']
-    assert_(isinstance(log_sp, SP.csc_matrix))
+    assert_(issparse(log_sp) and log_sp.format == "csc")
     assert_equal(log_sp.dtype.type, np.bool_)
     assert_array_equal(log_sp.toarray(),
                        [[True, True, True, False],
@@ -1178,10 +1173,17 @@ def test_empty_sparse():
     # Can we read empty sparse matrices?
     sio = BytesIO()
     import scipy.sparse
-    empty_sparse = scipy.sparse.csr_matrix([[0,0],[0,0]])
+    empty_sparse = scipy.sparse.csr_array([[0,0],[0,0]])
     savemat(sio, dict(x=empty_sparse))
     sio.seek(0)
-    res = loadmat(sio)
+
+    res = loadmat(sio, spmatrix=False)
+    assert not scipy.sparse.isspmatrix(res['x'])
+    res = loadmat(sio, spmatrix=True)
+    assert scipy.sparse.isspmatrix(res['x'])
+    res = loadmat(sio)  # chk default
+    assert scipy.sparse.isspmatrix(res['x'])
+
     assert_array_equal(res['x'].shape, empty_sparse.shape)
     assert_array_equal(res['x'].toarray(), 0)
     # Do empty sparse matrices get written with max nnz 1?
@@ -1239,7 +1241,7 @@ def test_save_unicode_field(tmpdir):
 
 def test_save_custom_array_type(tmpdir):
     class CustomArray:
-        def __array__(self):
+        def __array__(self, dtype=None, copy=None):
             return np.arange(6.0).reshape(2, 3)
     a = CustomArray()
     filename = os.path.join(str(tmpdir), 'test.mat')
@@ -1273,7 +1275,7 @@ def test_simplify_cells():
     (1, '8*_*', None),
 ])
 def test_matfile_version(version, filt, regex):
-    use_filt = pjoin(test_data_path, 'test*%s.mat' % filt)
+    use_filt = pjoin(test_data_path, f'test*{filt}.mat')
     files = glob(use_filt)
     if regex is not None:
         files = [file for file in files if re.match(regex, file) is not None]
@@ -1327,3 +1329,32 @@ def test_gh_17992(tmp_path):
             new_dict)
     assert_allclose(new_dict["data"][0][0], array_one)
     assert_allclose(new_dict["data"][0][1], array_two)
+
+
+def test_gh_19659(tmp_path):
+    d = {
+        "char_array": np.array([list("char"), list("char")], dtype="U1"),
+        "string_array": np.array(["string", "string"]),
+        }
+    outfile = tmp_path / "tmp.mat"
+    # should not error:
+    savemat(outfile, d, format="4")
+
+
+def test_large_m4():
+    # Test we can read a Matlab 4 file with array > 2GB.
+    # (In fact, test we get the correct error from reading a truncated
+    # version).
+    # See https://github.com/scipy/scipy/issues/21256
+    # Data file is first 1024 bytes of:
+    # >>> a = np.zeros((134217728, 3))
+    # >>> siom.savemat('big_m4.mat', {'a': a}, format='4')
+    truncated_mat = pjoin(test_data_path, 'debigged_m4.mat')
+    match = ("Not enough bytes to read matrix 'a';"
+             if np.intp == np.int64 else
+             "Variable 'a' has byte length longer than largest possible")
+    with pytest.raises(ValueError, match=match):
+        loadmat(truncated_mat)
+
+def test_gh_19223():
+    from scipy.io.matlab import varmats_from_mat  # noqa: F401
