@@ -1,6 +1,7 @@
 import numpy as np
-from numpy.testing import assert_array_almost_equal, assert_
-from scipy.sparse import csr_matrix, hstack
+from numpy.testing import assert_array_almost_equal, assert_, assert_array_equal
+from scipy.sparse import csr_matrix, csc_matrix, csr_array, csc_array, hstack
+from scipy import sparse
 import pytest
 
 
@@ -167,3 +168,47 @@ def test_csr_hstack_int64():
     X_hs_32 = hstack([X_1, X_3], format="csr")
     assert X_hs_32.indices.dtype == np.int32
     assert X_hs_32.indices.max() == max_int32 - 1
+
+@pytest.mark.parametrize("cls", [csr_matrix, csr_array, csc_matrix, csc_array])
+def test_mixed_index_dtype_int_indexing(cls):
+    # https://github.com/scipy/scipy/issues/20182
+    rng = np.random.default_rng(0)
+    base_mtx = cls(sparse.random(50, 50, random_state=rng, density=0.1))
+    indptr_64bit = base_mtx.copy()
+    indices_64bit = base_mtx.copy()
+    indptr_64bit.indptr = base_mtx.indptr.astype(np.int64)
+    indices_64bit.indices = base_mtx.indices.astype(np.int64)
+
+    for mtx in [base_mtx, indptr_64bit, indices_64bit]:
+        np.testing.assert_array_equal(
+            mtx[[1,2], :].toarray(),
+            base_mtx[[1, 2], :].toarray()
+        )
+        np.testing.assert_array_equal(
+            mtx[:, [1, 2]].toarray(),
+            base_mtx[:, [1, 2]].toarray()
+        )
+
+def test_broadcast_to():
+    a = np.array([1, 0, 2])
+    b = np.array([3])
+    e = np.zeros((0,))
+    res_a = csr_array(a)._broadcast_to((2,3))
+    res_b = csr_array(b)._broadcast_to((4,))
+    res_c = csr_array(b)._broadcast_to((2,4))
+    res_d = csr_array(b)._broadcast_to((1,))
+    res_e = csr_array(e)._broadcast_to((4,0))
+    assert_array_equal(res_a.toarray(), np.broadcast_to(a, (2,3)))
+    assert_array_equal(res_b.toarray(), np.broadcast_to(b, (4,)))
+    assert_array_equal(res_c.toarray(), np.broadcast_to(b, (2,4)))
+    assert_array_equal(res_d.toarray(), np.broadcast_to(b, (1,)))
+    assert_array_equal(res_e.toarray(), np.broadcast_to(e, (4,0)))
+
+    with pytest.raises(ValueError, match="cannot be broadcast"):
+        csr_matrix([[1, 2, 0], [3, 0, 1]])._broadcast_to(shape=(2, 1))
+
+    with pytest.raises(ValueError, match="cannot be broadcast"):
+        csr_matrix([[0, 1, 2]])._broadcast_to(shape=(3, 2))
+
+    with pytest.raises(ValueError, match="cannot be broadcast"):
+        csr_array([0, 1, 2])._broadcast_to(shape=(3, 2))
